def find_all(arr,x):
    if not isinstance(arr, list):
        raise TypeError(f"Expected a list, but got {type(arr).__name__}")
    indices=[]
    for i in range(len(arr)):
        if arr[i]==x:
            indices.append(i)
    return indices

def find_nth(arr,x,n):
    if not isinstance(arr, list):
        raise TypeError(f"Expected a list, but got {type(arr).__name__}")
    found=0
    for i in range(len(arr)):
        if arr[i]==x:
            found+=1
            if found==n:
                return i
    return -1

def find_unique(arr):
    if not isinstance(arr,list):
        raise TypeError(f"Expected a list, but got {type(arr).__name__}")
    return list(set(arr))

def find_nth_highest(arr,n):
    if not isinstance(arr,list):
        raise TypeError(f"Expected a list, but got {type(arr).__name__}")
    return sorted(arr)[(-1*n)]

def find_nth_lowest(arr,n):
    if not isinstance(arr,list):
        raise TypeError(f"Expected a list, but got {type(arr).__name__}")
    return sorted(arr)[(n-1)]

def find_intersections(arr1,arr2):
    if not isinstance(arr1,list) and not isinstance(arr2,list):
        raise TypeError(f"Expected lists, but got {type(arr1).__name__}, {type(arr2).__name__}")
    return list(set(arr1) & set(arr2))


import inits
import math

def flatten(arr):
    if not isinstance(arr,list):
        raise TypeError(f"Expected a list, but got {type(arr).__name__}")
    a=[]
    for element in arr:
        if isinstance(element,list):
            a.extend(flatten(element))
        else:
            a.append(element)
    return a

def sliding_window(arr,win_size):
    if not isinstance(arr,list):
        raise TypeError(f"Expected a list, but got {type(arr).__name__}")
    return list(arr[i:i+win_size] for i in range(len(arr)-win_size+1))

def reverse(arr):
    if not isinstance(arr,list):
        raise TypeError(f"Expected a list, but got {type(arr).__name__}")
    return arr[::-1]

def concatenate(arr) -> str:
    if not isinstance(arr,list):
        raise TypeError(f"Expected a list, but got {type(arr).__name__}")
    a = ''
    for element in arr:
        if isinstance(element, list):
            a+=str(concatenate(element))
        else:
            a+=str(element)
    return a

def chunk(arr,chunk_size=-1,chunk_quantity=-1): #Could be optimized to work with non-factors
    if not isinstance(arr,list):
        raise TypeError(f"Expected a list, but got {type(arr).__name__}")
    if (chunk_size!=-1 and not chunk_quantity==-1) or (chunk_size==-1 and not chunk_quantity!=-1):
        raise ValueError("Must define only either chunk size or chunk quantity")
    chunks=[]
    if chunk_size!=-1:
        for i in range(math.ceil(len(arr)/chunk_size)):
            chunks.append(arr[i*chunk_size:(i+1)*chunk_size])
        return chunks
    if chunk_quantity!=-1:
        for i in range(chunk_quantity):
            chunks.append(arr[math.ceil(len(arr)/chunk_quantity)*i:(i + 1) * math.ceil(len(arr)/chunk_quantity)])
        return chunks

def remove_duplicates(arr):
    if not isinstance(arr,list):
        raise TypeError(f"Expected a list, but got {type(arr).__name__}")
    return list(dict().fromkeys(arr))

def zip_arrays(arr1,arr2):
    if not isinstance(arr1,list) and not isinstance(arr2,list):
        raise TypeError(f"Expected lists, but got {type(arr1).__name__}, {type(arr2).__name__}")
    return list(zip(arr1,arr2))


import inits
def selectionSort(arr: list,display=False):
    if not isinstance(arr,list):
        raise TypeError(f"Expected a list, but got {type(arr).__name__}")
    for i in range(len(arr)-1):
        min_index=i
        for j in range(i+1,len(arr)):
            if arr[min_index]>arr[j]:
                min_index=j
        if display:
            inits.chart(arr)
        if i!=min_index:
            arr[i], arr[min_index] = arr[min_index], arr[i]
    return arr

def bubbleSort(arr: list,display=False):
    if not isinstance(arr,list):
        raise TypeError(f"Expected a list, but got {type(arr).__name__}")
    for i in range(len(arr)):
        swapped=False
        for j in range(0,len(arr)-i-1):
            if arr[j]>arr[j+1]:
                arr[j+1],arr[j]=arr[j],arr[j+1]
                swapped=True
        if display:
            inits.chart(arr)
        if not swapped:
            break
    return arr

def insertionSort(arr: list,display=False):
    if not isinstance(arr,list):
        raise TypeError(f"Expected a list, but got {type(arr).__name__}")
    for i in range(1,len(arr)):
        key=arr[i]
        j=i-1
        while j>=0 and key<arr[j]:
            arr[j+1]=arr[j]
            j-=1
        arr[j+1]=key
        if display:
            inits.chart(arr)

    return arr

def merge(arr, left, mid, right,display):
    n1 = mid - left + 1
    n2 = right - mid
    L = [0] * n1
    R = [0] * n2
    for i in range(n1):
        L[i] = arr[left + i]
    for j in range(n2):
        R[j] = arr[mid + 1 + j]
    i = 0
    j = 0
    k = left
    while i < n1 and j < n2:
        if L[i] <= R[j]:
            arr[k] = L[i]
            i += 1
        else:
            arr[k] = R[j]
            j += 1
        k += 1
    while i < n1:
        arr[k] = L[i]
        i += 1
        k += 1
    while j < n2:
        arr[k] = R[j]
        j += 1
        k += 1
    if display:
        inits.chart(arr)
def mergeSort(arr,left,right,display=False):
    if not isinstance(arr,list):
        raise TypeError(f"Expected a list, but got {type(arr).__name__}")
    if left < right:
        mid = (left + right) // 2

        mergeSort(arr, left, mid,display)
        mergeSort(arr, mid + 1, right,display)
        merge(arr, left, mid, right,display)
    return arr


import random
from transformations import concatenate
import matplotlib.pyplot as plt

def empty(dimensions:list[int], datatype: type):
    if not isinstance(dimensions, list):
        raise TypeError(f"Expected type 'list[int]', but got type {type(dimensions).__name__}")
    else:
        if not all(isinstance(d, int) for d in dimensions):
            raise TypeError(f"Expected type 'list[int]', but got some non-int value")
    if not isinstance(datatype,type):
        raise TypeError(f"Expected type 'type', but got {type(datatype).__name__}")
    if len(dimensions) == 1:
        return [datatype() for _ in range(dimensions[0])]
    else:
        return [empty(dimensions[1:], datatype) for _ in range(dimensions[0])]

def uniform_fill(dimensions:list[int], value):
    if not isinstance(dimensions, list):
        raise TypeError(f"Expected type 'list[int]', but got type {type(dimensions).__name__}")
    else:
        if not all(isinstance(d, int) for d in dimensions):
            raise TypeError(f"Expected type 'list[int]', but got some non-int value")
    if len(dimensions) == 1:
        return [value for _ in range(dimensions[0])]
    else:
        return [uniform_fill(dimensions[1:], value) for _ in range(dimensions[0])]

def randoms(length,datatype,minimum=-100,maximum=100):
    arr=[]
    if datatype in [int,float,complex]:
        for _ in range(length):
            arr.append(random.randint(minimum,maximum))
    if datatype==str:
        for _ in range(length):
            st=''
            for _ in range(random.randint(1,10)):
                st+=random.choice([chr(random.randint(65,90)),chr(random.randint(97,122))])
            arr.append(st)
    if datatype==bool:
        for _ in range(length):
            arr.append(random.choice([True,False]))
    if datatype==dict:
        for _ in range(length):
            arr.append({concatenate(randoms(1,str)):str(random.randint(-100,100))})
    if datatype==list:
        for _ in range(length):
            x=[]
            for _ in range(random.randint(1, 10)):
                x.append(concatenate(randoms(1,str)))
            arr.append(x)
    if datatype==tuple:
        for _ in range(length):
            x=[]
            for _ in range(random.randint(2, 5)):
                x.append(concatenate(randoms(1,int)))
            arr.append(tuple(x))
    if datatype==range:
        for _ in range(length):
            for _ in range(2):
                arr.append(range(random.randrange(0,100)))
    if datatype==set:
        for _ in range(length):
            arr.append({concatenate(randoms(1, str)): str(random.randint(-100, 100))})
    return arr

def chart(arr):
    plt.ion()
    plt.clf()
    graph = plt.plot(list(range(len(arr))), arr)
    plt.pause(0.000001)
